/*
 * lomoco list -- a doubly-linked list
 *
 * This code is based on glist.{h,c} from glib
 *   ftp://ftp.gtk.org/pub/gtk/
 * Copyright (c) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (c) 2006 by Andreas Schneider <mail@lomocoapses.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * vim: ts=2 sw=2 et cindent
 */

#include <stdio.h>
#include <stdlib.h>

#include "lomoco_list.h"

/*
 * Adds a new element on to the end of the list.
 */
lomoco_list_t *lomoco_list_append(lomoco_list_t *list, lomoco_pointer data) {
  lomoco_list_t *new, *last;

  new = lomoco_list_alloc();
  new->data = data;

  if (list) {
    last = lomoco_list_last(list);

    last->next = new;
    new->prev = last;

    return list;
  } else {
    return new;
  }
}

/*
 * Adds a new element on at the beginning of the list.
 */
lomoco_list_t *lomoco_list_prepend(lomoco_list_t *list, lomoco_pointer data) {
  lomoco_list_t *new, *first;

  new = lomoco_list_alloc();
  new->data = data;

  if (list) {
    first = lomoco_list_first(list);

    first->prev = new;
    new->next = first;
  }

  return new;
}

/*
 * Inserts a new element into the list at the given position.
 */
lomoco_list_t *lomoco_list_insert(lomoco_list_t *list, lomoco_pointer data, long position) {
  lomoco_list_t *new, *temp;

  /* Handle wrong values for position */
  if (position < 0) {
    return lomoco_list_append (list, data);
  } else if (position == 0) {
    return lomoco_list_prepend (list, data);
  }

  temp = lomoco_list_position(list, position);

  if (!temp) {
    return lomoco_list_append(list, data);
  }

  new = lomoco_list_alloc();
  new->data = data;

  /* List is not empty */
  if (temp->prev) {
    temp->prev->next = new;
    new->prev = temp->prev;
  }

  new->next = temp;
  temp->prev = new;

  /*  */
  if (temp == list) {
    return new;
  } else {
    return list;
  }
}

/*
 * Inserts a new element into the list, using the given comparison function to
 * determine its position.
 */
lomoco_list_t *lomoco_list_insert_sorted(lomoco_list_t *list, lomoco_pointer data, lomoco_compare_func func) {
  lomoco_list_t *new, *temp;
  int cmp;

  new = lomoco_list_alloc();
  new->data = data;

  /* list is empty */
  if (!list) {
    return new;
  }

  temp = list;
  cmp = ((lomoco_compare_func) func) (data, temp->data);

  while ((temp->next) && (cmp > 0)) {
    temp = temp->next;

    cmp = ((lomoco_compare_func) func) (data, temp->data);
  }

  /* last element */
  if ((!temp->next) && (cmp > 0)) {
    temp->next = new;
    new->prev = temp;
    return list;
  }

  /* first element */
  if (temp->prev) {
    temp->prev->next = new;
    new->prev = temp->prev;
  }

  new->next = temp;
  temp->prev = new;

  /* inserted before first */
  if (temp == list) {
    return new;
  } else {
    return list;
  }
}

/*
 * Allocates space for one lomoco_list_t element.
 */
lomoco_list_t *lomoco_list_alloc(void) {
  lomoco_list_t *list;

  list = calloc(1, sizeof(lomoco_list_t));
  if (list == NULL) {
    return NULL;
  }

  list->data = NULL;

  list->prev = NULL;
  list->next = NULL;

  return list;
}

/*
 * Removes an element from a lomoco_list. If two elements contain the same data,
 * only the first is removed.
 */
lomoco_list_t *lomoco_list_remove(lomoco_list_t *list, lomoco_pointer data) {
  lomoco_list_t *temp;

  temp = list;

  while (temp) {
    if (temp->data != data) {
      temp = temp->next;
    } else {
      /* not at first element */
      if (temp->prev) {
        temp->prev->next = temp->next;
      }

      /* not at last element */
      if (temp->next) {
        temp->next->prev = temp->prev;
      }

      /* first element */
      if (list == temp) {
        list = list->next;
      }

      free (temp);
      break;
    }
  }

  return list;
}

/*
 * Frees all elements from a lomoco_list.
 */
void lomoco_list_free(lomoco_list_t *list) {
  lomoco_list_t *temp;
  if (list) {
    list = lomoco_list_last(list);

    while (list->prev) {
      temp = list;
      list = list->prev;

      free (temp);
    }
  }
}

/*
 * Gets the next element in a lomoco_list.
 */
lomoco_list_t *lomoco_list_next(lomoco_list_t *list) {
  if (list) {
    return list->next;
  } else {
    return NULL;
  }
}

/*
 * Gets the previous element in a lomoco_list.
 */
lomoco_list_t *lomoco_list_previous(lomoco_list_t *list) {
  if (list) {
    return list->prev;
  } else {
    return NULL;
  }
}

/*
 * Gets the number of elements in a lomoco_list
 */
unsigned long lomoco_list_length(lomoco_list_t *list) {
  unsigned long length = 0;

  if (list) {
    while (list->next) {
      length++;
      list = list->next;
    }
  }

  return length;
}

/*
 * Gets the first element in a lomoco_list
 */
lomoco_list_t *lomoco_list_first(lomoco_list_t *list) {

  if (list) {
    while (list->prev) {
      list = list->prev;
    }
  }

  return list;
}

/*
 * Gets the last element in a lomoco_list
 */
lomoco_list_t *lomoco_list_last(lomoco_list_t *list) {

  if (list) {
    while (list->next) {
      list = list->next;
    }
  }

  return list;
}

/*
 * Gets the element at the given positon in a lomoco_list
 */
lomoco_list_t *lomoco_list_position(lomoco_list_t *list, long position) {

  while ((position-- > 0) && list) {
    list = list->next;
  }

  return list;
}

/*
 * Finds the element in a lomoco_list_t which contains the given data.
 */
lomoco_list_t *lomoco_list_find(lomoco_list_t *list, lomoco_pointer data) {
  if (list) {
    while (list) {
      if (list->data == data) {
        break;
      }
      list = list->next;
    }
    return list;
  }

  return NULL;
}

/*
 * Finds an element, using a supplied function to find the desired
 * element.
 */
lomoco_list_t *lomoco_list_find_custom(lomoco_list_t *list, lomoco_pointer data, lomoco_compare_func func) {
  int cmp;

  if (list && func) {
    while (list) {
      cmp = ((lomoco_compare_func) func) (list->data, data);
      if (cmp == 0) {
        return list;
      }
      list = list->next;
    }
  }

  return NULL;
}

/*
 * Sorts the elements of a lomoco_list.
 */
lomoco_list_t *lomoco_list_sort(lomoco_list_t *list, lomoco_compare_func func) {
  lomoco_list_t *second;
  /* list is empty */
  if (list == NULL) {
    return NULL;
  /* list has only one element */
  } else if (list->next == NULL) {
    return list;
  } else {
    /* split list */
    second = lomoco_list_split(list);
    return lomoco_list_merge(lomoco_list_sort(list, func), lomoco_list_sort(second, func), func);
  }
}

/*
 * Internal used function to merge 2 lists using a compare function
 */
lomoco_list_t *lomoco_list_merge(lomoco_list_t *list1, lomoco_list_t *list2, lomoco_compare_func func) {
  int cmp;

  /* lists are emty */
  if (list1 == NULL) {
    return list2;
  } else if (list2 == NULL) {
    return list1;
  }

  cmp = ((lomoco_compare_func) func)(list1->data, list2->data);
  /* compare if it is smaller */
  if (cmp <= 0) {
    list1->next = lomoco_list_merge(list1->next, list2, func);
    if (list1->next) {
      list1->next->prev = list1;
    }return list1;
  } else {
    list2->next = lomoco_list_merge(list1, list2->next, func);
    if (list2->next) {
      list2->next->prev = list2;
    }
    return list2;
  }
}

/*
 * Internally used function to split 2 lists.
 */
lomoco_list_t *lomoco_list_split(lomoco_list_t *list) {
  lomoco_list_t *second;

  /* list is empty */
  if (list == NULL) {
    return NULL;

  /* list has only 1 element */
  } else if (list->next == NULL) {
    return NULL;

  /* split */
  } else {
    second = list->next;
    list->next = second->next;
    /* is last element */
    if (list->next) {
      list->next->prev = list;
    }

    second->prev = NULL;
    second->next = lomoco_list_split(second->next);
    /* is last element */
    if (second->next) {
      second->next->prev = second;
    }
    return second;
  }
}

