/*
 * lomoco -- a doubly-linked list
 *
 * This code is based on glist.{h,c} from glib
 *   ftp://ftp.gtk.org/pub/gtk/
 * Copyright (c) 1995-1997  Peter Mattis, Spencer Kimball and Josh MacDonald
 * Copyright (c) 2006 by Andreas Schneider <mail@lomocoapses.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * vim: ts=2 sw=2 et cindent
 */

#ifndef LOMOCO_LIST_H
#define LOMOCO_LIST_H

/**
 * lomoco_list_t -- a doubly-linked list.
 *
 * The lomoco_list_t structure and its associated functions provide a standard
 * doubly-linked list data structure. Each node has two links: one points to
 * the previous node, or points to a null value or empty list if it is the
 * first  node; and one points to the next, or points to a null value or empty
 * list if it is the final node.
 *
 * The data contained in each element can be simply pointers to any type of
 * data. You are the owner of the data, this means you have to free the memory
 * you have allocated for the data.
 *
 * @file   lomoco_list.h
 */

/**
 * An untyped pointer.
 */
typedef void *lomoco_pointer;

/**
 * An untyped pointer to constant data. The data pointed to should not be
 * changed.
 *
 * This is typically used in function prototypes to indicate that the data
 * pointed to will not be altered by the function.
 */
typedef const void *lomoco_const_pointer;

/**
 * Specifies the type of a comparison function used to compare two values. The
 * value which should be returned depends on the context in which the
 * lomoco_compare_func is used.
 *
 * @param a             First parameter to compare with.
 *
 * @param b             Second parameter to compare with.
 *
 * @return              The function should return a number > 0 if the first
 *                      parameter comes after the second parameter in the sort
 *                      order.
 */
typedef int (*lomoco_compare_func) (lomoco_const_pointer a, lomoco_const_pointer b);

/**
 * Used for each element in a doubly-linked list. The data field holds the
 * element's data, which can be a pointer to any kind of data.
 * The next and prev pointers are the links to the next and previous elements
 * in the list.
 */
typedef struct lomoco_list_s {
  struct lomoco_list_s *next;
  struct lomoco_list_s *prev;

  lomoco_pointer data;
} lomoco_list_t;


/**
 * Adds a new element on to the end of the list.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data for the new element.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_append(lomoco_list_t *list, lomoco_pointer data);

/**
 * Adds a new element on at the beginning of the list.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data for the new element.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_prepend(lomoco_list_t *list, lomoco_pointer data);

/**
 * Inserts a new element into the list at the given position. If the position
 * is lesser than 0, the new element gets appended to the list, if the position
 * is 0, we prepend the element and if the given position is greater than the
 * length of the list, the element gets appended too.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data for the new element.
 *
 * @param position      The position to insert the element.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_insert(lomoco_list_t *list, lomoco_pointer data, long position);

/**
 * Inserts a new element into the list, using the given comparison function to
 * determine its position.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data for the new element.
 *
 * @param func          The function to compare elements in the list. It
 *                      should return a number > 0 if the first parameter comes
 *                      after the second parameter in the sort order.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_insert_sorted(lomoco_list_t *list, lomoco_pointer data, lomoco_compare_func func);

/**
 * Allocates space for one lomoco_list_t element.
 *
 * @return             A pointer to the newly-allocated element.
 */
lomoco_list_t *lomoco_list_alloc(void);

/**
 * Removes an element from a lomoco_list. If two elements contain the same data,
 * only the first is removed.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data of the element to remove.
 */
lomoco_list_t *lomoco_list_remove(lomoco_list_t *list, lomoco_pointer data);

/**
 * Frees all elements from a lomoco_list.
 *
 * @param list          A pointer to lomoco_list.
 */
void lomoco_list_free(lomoco_list_t *list);

/**
 * Gets the previous element in a lomoco_list.
 *
 * @param               An element in a lomoco_list.
 *
 * @return              The previous element, or NULL if there are no more
 *                      elements.
 */
lomoco_list_t *lomoco_list_previous(lomoco_list_t *list);

/**
 * Gets the next element in a lomoco_list.
 *
 * @param               An element in a lomoco_list.
 *
 * @return              The next element, or NULL if there are no more
 *                      elements.
 */
lomoco_list_t *lomoco_list_next(lomoco_list_t *list);

/**
 * Gets the number of elements in a lomoco_list
 *
 * @param list          A pointer to lomoco_list.
 *
 * @return              The number of elements
 */
unsigned long lomoco_list_length(lomoco_list_t *list);

/**
 * Gets the first element in a lomoco_list
 *
 * @param list          A pointer to lomoco_list.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_first(lomoco_list_t *list);

/**
 * Gets the last element in a lomoco_list
 *
 * @param list          A pointer to lomoco_list.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_last(lomoco_list_t *list);

/**
 * Gets the element at the given positon in a lomoco_list.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param position      The position of the element, counting from 0.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_position(lomoco_list_t *list, long position);

/**
 * Finds the element in a lomoco_list_t which contains the given data.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data of the element to remove.
 *
 * @return              The found element or NULL if it is not found.
 */
lomoco_list_t *lomoco_list_find(lomoco_list_t *list, lomoco_pointer data);

/**
 * Finds an element, using a supplied function to find the desired
 * element.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param data          The data of the element to remove.
 *
 * @param func          The function to call for each element. It should
 *                      return 0 when the desired element is found.
 *
 * @return              The found element or NULL if it is not found.
 */
lomoco_list_t *lomoco_list_find_custom(lomoco_list_t *list, lomoco_pointer data, lomoco_compare_func func);

/**
 * Sorts the elements of a lomoco_list.
 * The algorithm used is Mergesort, because that works really well
 * on linked lists, without requiring the O(N) extra space it needs
 * when you do it on arrays.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @param func          The comparison function used to sort the lomoco_list. This
 *                      function is passed 2 elements of the GList and should
 *                      return 0 if they are equal, a negative value if the first
 *                      element comes before the second, or a positive value if the
 *                      first element comes after the second.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_sort(lomoco_list_t *list, lomoco_compare_func func);

/**
 * Internal used function to merge 2 lists using a compare function
 *
 * @param list1         A pointer to lomoco_list.
 *
 * @param list2         A pointer to lomoco_list.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_merge(lomoco_list_t *list1, lomoco_list_t *list2, lomoco_compare_func func);

/**
 * Internally used function to split 2 lists.
 *
 * @param list          A pointer to lomoco_list.
 *
 * @return              New start of the list, which may have changed, so make
 *                      sure you store the new value.
 */
lomoco_list_t *lomoco_list_split(lomoco_list_t *list);

#endif /* LOMOCO_LIST_H */

