/*
 * lomoco - Logitech Mouse Control for Linux
 *
 * Copyright (c) 2006 by Andreas Schneider <mail@cynapses.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 * vim: ts=2 sw=2 et cindent
 */

#include <errno.h>
#include <stdio.h>
#include <usb.h>

#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/* you have to inlcude asm/types.h before hiddev */
#include <asm/types.h>
#include <linux/hiddev.h>

#include "lomoco_debug.h"
#include "lomoco_scan.h"
#include "lomoco_list.h"

#include "logitech/lomoco_logitech.h"

lomoco_list_t *lomoco_get_usb_devices(lomoco_base_t *base, lomoco_list_t *list) {
  struct usb_bus *usb_bus;
  struct usb_device *usb_device;
  usb_dev_handle *usb_handle;

  lomoco_device_t *lomoco_device;

  /* init usb */
  usb_init();

  /* FIXME: more than 10? */
  /* set libusb debug level */
  if (lomoco_get_debug_level() >= 10) {
    usb_set_debug(1);
  }

  if (usb_find_busses() < 0) {
    /* TODO: error handling */
    return NULL;
  }

  if (usb_find_devices() < 0) {
    /* TODO: error handling */
    return NULL;
  }

  /* scan bus */
  for (usb_bus = usb_busses; usb_bus != NULL; usb_bus = usb_bus->next) {
    for (usb_device = usb_bus->devices; usb_device != NULL; usb_device = usb_device->next) {

      /* we have a logitech mouse */
      if (usb_device->descriptor.idVendor == base->logitech_vid) {

        lomoco_device = lomoco_get_logitech_usb_device(base, usb_device->descriptor.idProduct);

        /* add it to the list */
        if (lomoco_device != NULL) {
          list = lomoco_list_append(list, lomoco_device);
        }

      }
    }
  }

  return list;
}

lomoco_list_t *lomoco_get_hid_devices(lomoco_base_t *base, lomoco_list_t *list) {

  lomoco_device_t *lomoco_device;
  int i, fd = -1;
  char evdevname[32] = {0};

  /*
   * ioctl() requires a file descriptor, so we check we got one, and
   * then open it
   */
  for (i = 0; i < 64; i++) {
    struct hiddev_devinfo device_info;

    sprintf(evdevname, "/dev/hiddev%d", i);

    /* try to open the device */
    if ((fd = open(evdevname, O_RDONLY)) < 0) {
      LOMOCO_DEBUG(12, "%s couldn't be opened", evdevname);

      /* try other path */
      sprintf(evdevname, "/dev/usb/hiddev%d", i);

      if ((fd = open(evdevname, O_RDONLY)) < 0) {
        LOMOCO_DEBUG(12, "%s couldn't be opened", evdevname);
        continue;
      }
    }
    LOMOCO_DEBUG(9, "hiddev event device is %s", evdevname);

    /* get device information */
    if (ioctl(fd, HIDIOCGDEVINFO, &device_info) < 0) {
      LOMOCO_ERRNO(errno, "hiddev information couldn't be retrieved");
      continue;
    } else {
      LOMOCO_DEBUG(9, "hiddev device found with vid: 0x%04hx (%d), pid: 0x%04hx (%d)",
          device_info.vendor, device_info.vendor, device_info.product,
          device_info.product); }

    /* we have a logitech mouse */
    if ((int)(unsigned short) device_info.vendor == base->logitech_vid) {

      lomoco_device = lomoco_get_logitech_hid_device(base, (int)(unsigned short) device_info.product);

      /* add it to the list */
      if (lomoco_device != NULL) {
        list = lomoco_list_append(list, lomoco_device);
      }
    }
    close(fd);
  }

  return list;
}

